<?php
/******************************************************************************
Filename:   rt2mei.php
Version:    1.1a
Date:       4 March 2009
Author:     Wooders
Website:    http://www.wooders.co.uk/rt2mei

Preconditions:
This script requires PHP 4.3.5 or greater to run.

Purpose:
To read EPG data from the Radio Times website and write it into a file
MyStuff.mei.

Optionally, the channels selected in the configuration file will be written
in the order specified into a Favourites file, named rt2mei.mcl.

These files may then be transferred to the Toppy and read by the MyStuff TAP,
which will then re-populate its EPG with this data.

It is the user's responsibility to transfer the files onto the Toppy.

The format of the output files will be correct when this script is run on
Windows, Linux and Mac. It has been known for other operating systems to
produce wrong results so please check that the line terminator is CRLF.

-------------------------------------------------------------------------------

All downloaded data is the copyright of the Radio Times website
(http://www.radiotimes.com). Use of this data is restricted to personal use
only.

-------------------------------------------------------------------------------

Simple Instructions:

Assuming that

a) PHP is in the PATH and
b) that both rt2mei.php and rt2mei.cfg have been copied into the directory
   /public/rt2mei

Type the following at a command line

php /public/rt2mei/rt2mei.php

MyStuff.mei and rt2mei.mcl both will be written into /public/rt2mei

-------------------------------------------------------------------------------

Usage:
Assuming that PHP is in the PATH then

php <path>rt2mei.php [-- options]

Where options are:
    -h      prints this information

    -q      quiet mode, no error messages output.
	
    -f      suppresses the writing of a Favourites file

    -r      replace ':' in title with '_'

    -i      followed by a path to the directory where the rt2mei.cfg
            configuration file is located. If omitted then the directory path
            for this script is used.

    -o      followed by a path to the directory where MyStuff.mei and
            rt2mei.mcl will be written. If omitted then the directory
            path for this script is used.

    -w      followed by the full path to the wget executable forces wget to be
            used to grab the EPG data. This is only normally necessary if you
            are behind a corporate firewall or similar. wget is not included in
            rt2mei.

    -t      followed by the full path to the directory where temporary files
            should be written. This is only needed when the -w switch is used.

    -new    rejects EPG data if it contains a program that ends before 0000 hrs
            today

    -mt     does not reject empty EPG data files. I recommend that you DO NOT
            use this option because you may lose EPG data. If you insist on
            using it then be it on your own head, don't come whining to me!

	-d      followed by the full path to the directory where the raw radio
            times data files should be cached for later use by the -s option.

	-s      followed by the URL or directory where the raw radio times data
            files should be read from.

			
This allows rt2mei.php, rt2mei.cfg and MyStuff.mei all to be in different
directories.

User Includes

rt2mei now features user included code. Include statements have been added at
strategic points in the code to allow the user to add his own code to customise
the output. The following files may be included:

Filename:   rt2mei_character_replacements.php
NOTE: This is not an optional file, it must exist. If you don't understand it
then don't touch it.
This code will be executed for every line of program data downloaded from the
RT website. It is executed immediately before each field in the line has been
written into its own variable. This is the place to filter or modify any
strange characters. Default replacement of certain characters has been included
but improvements are possible, especially if you have iconv() available.

Filename:   user_epg_line_fields_split.php
This is an optional file; it does not have to exist.
This code will be executed for every line of program data downloaded from the
RT website. It is executed immediately after each field in the line has been
written into its own variable. No processing or alteration of the variables by
rt2mei has yet taken place.

Filename:   user_pre_mei_line.php
This is an optional file; it does not have to exist.
This code will be executed for every line of program data downloaded from the
RT website. It is executed immediately before the line of data to be written
to the mei file is constructed.

Filename:   user_pre_cfg_read.php
This is an optional file; it does not have to exist.
This code will be executed once every time rt2mei.php is run. It is executed
just prior to the rt2mei.cfg file being read. If you want to do something
before rt2mei has done very much then this would be a good place to do it.

Filename:   user_pre_mei_create.php
This is an optional file; it does not have to exist.
This code will be executed just after the rt2mei.cfg file is processed and the
Favourites file is written (optional), and just before the mei file is written.

Filename:   user_pre_channel_epg_read.php
This is an optional file; it does not have to exist.
This code will be executed once for every file of program data downloaded from
the RT website. It is executed immediately after the file has been read and
before its lines of data are processed.

Filename:   user_post_channel_epg_read.php
This is an optional file; it does not have to exist.
This code will be executed once for every file of program data downloaded from
the RT website. It is executed immediately after the lines of data in the file
have been processed.

Filename:   user_success.php
This is an optional file; it does not have to exist.
This code will be executed if the complete process was completed successfully,
i.e. all files of program data read and processed with no errors.


I recommend the following regarding include files:

1) Do NOT delete rt2mei_character_replacements.php

2) Only create those include files that you need.

3) Writing your own code could affect the functionality of rt2mei so only do it
   if you are competent.

4) Choose variable names that are different to those used in rt2mei.

5) Use the lines below as a template for your include files

---- copy the lines below this one ----
<?php
// It is useful to retain the following line of code especially when debugging
if (!$quiet) echo __FILE__ . " included.\n\n";  // Show that this include file has run


// Now include your own code below this line ...


// ... and above this one.
?>
---- copy all lines up to but excluding this one ----


The following list of variables may prove helpful

*   $rt_title               program title
*   $rt_sub_title           secondary title, e.g. episode title
*   $rt_episode             episode, either blank or [this episode number]/[total number of episodes in this series], e.g. 2/6
*   $rt_year                production date, blank or 4-digit year
    $rt_director            not used by rt2mei, blank or director's name
    $rt_cast                not used by rt2mei, blank or a list of cast members
*   $rt_premiere            'true' if it is a premiere else 'false'
    $rt_film                'true' if it is a film else 'false'
*   $rt_repeat              'true' if it is a repeat else 'false'
*   $rt_subtitles           'true' if it is subtitled else 'false'
*   $rt_widescreen          'true' if it is in widescreen else 'false'
    $rt_new_series          'true' if it is a new series else 'false'
    $rt_deaf_signed         'true' if it is signed for the deaf else 'false'
*   $rt_black_and_white     'true' if it is broadcast in B&W else 'false'
*   $rt_star_rating         a single digit between 1 and 5
*   $rt_certificate         blank, PG, 12, 15 or 18
*   $rt_genre               description of genre
*   $rt_desc                description of the episode or program
    $rt_choice              not used by rt2mei
    $rt_date                date on which the program start is shown as dd/mm/yyy
    $rt_start               time at which program starts is shown as hh:mm
    $rt_stop                time at which program stops is shown as hh:mm
*   $rt_duration_mins       program length in minutes

The variables marked with a * will change the MEI data if they are changed.


Enjoy

  Wooders

Acknowledgements

BobD,       http://www.bobdsmystuff.co.uk
            BobD is the author of the MyStuff TAP for the Topfield 5800 PVR
            that reads the MyStuff.mei file in order to populate its EPG.

Bawbagg,    http://my.opera.com/bawbagg/
            Bawbagg has written the MyStuff Extended Info application that
            runs under Windows. My thanks to him for his co-operation and
            support in producing this alternative method of generating a
            MyStuff.mei file

Ed Avis,    ed@membled.com
            Ed is the author of the tv_grab_uk_rt component of XMLTV that
            helped me decode the Radio Times datastream. Thanks Ed.

******************************************************************************/


class MEI
{
    var $filename = '';
    var $handle = false;
    var $rt_id = 0;
    var $lcn = 0;
    var $rt_channel_name = '';
    var $svc_limited = false;
    var $svc_start_hrs = 0;
    var $svc_start_mins = 0;
    var $svc_stop_hrs = 0;
    var $svc_stop_mins = 0;
    var $tstamp;
    var $quiet;
    var $replacebadchars;
    var $rt2mei_character_replacements_path;
    var $rt2mei_character_replacements_exists;
    var $user_epg_line_fields_split_path;
    var $user_epg_line_fields_split_exists;
    var $user_pre_mei_line_path;
    var $user_pre_mei_line_exists;

    // Constructor
    function MEI($filename, $tstamp, $quiet, $replacebadchars)
    {
        $this->filename = $filename;
        $this->tstamp = $tstamp;
        $this->quiet = $quiet;
        $this->replacebadchars = $replacebadchars;
        $this->handle = @fopen($this->filename, 'w');
        if ($this->handle === false)
        {
            if (!$this->$quiet)
                echo "Failed to open $this->filename\n";
            exit(1);
        }
    }

    // Destructor
    function destructor()
    {
        if ($this->handle != false)
        {
            fclose($this->handle);
            $this->handle = false;
        }
    }

    function set_rt_id($val)
    {
        $this->rt_id = $val;
    }

    function set_lcn($val)
    {
        $this->lcn = $val;
    }

    function set_rt_channel_name($val)
    {
        $this->rt_channel_name = $val;
    }

    function set_svc_hours($svc_limited, $svc_start, $svc_stop)
    {
        $this->svc_limited = $svc_limited;
        if ($svc_limited == true)
        {
            $this->svc_start_hrs = (int) substr($svc_start, 0, 2);
            $this->svc_start_mins = (int) substr($svc_start, -2);
            $this->svc_stop_hrs = (int) substr($svc_stop, 0, 2);
            $this->svc_stop_mins = (int) substr($svc_stop, -2);
        }
    }

    function set_script_files($rt2mei_character_replacements_path, $rt2mei_character_replacements_exists,
                              $user_epg_line_fields_split_path, $user_epg_line_fields_split_exists,
                              $user_pre_mei_line_path, $user_pre_mei_line_exists)
    {
        $this->rt2mei_character_replacements_path = $rt2mei_character_replacements_path;
        $this->rt2mei_character_replacements_exists = $rt2mei_character_replacements_exists;
        $this->user_epg_line_fields_split_path = $user_epg_line_fields_split_path;
        $this->user_epg_line_fields_split_exists = $user_epg_line_fields_split_exists;
        $this->user_pre_mei_line_path = $user_pre_mei_line_path;
        $this->user_pre_mei_line_exists = $user_pre_mei_line_exists;
    }

    function process_line(&$line, $newepg)
    {
        $quiet = $this->quiet;                  // all include files may refer to $quiet

        $fields = explode("~", $line);          // split the line into its various fields
        
        if (count($fields) < 23)
            return(false);                      // the line does not hold any identifiable fields so ignore the line

        foreach ($fields as $key => $value)
        {
            if (is_null($value))
                $value = "";
        }

        if ($this->rt2mei_character_replacements_exists)
            include $this->rt2mei_character_replacements_path;

        // Write each field into its own variable
        list($rt_title, $rt_sub_title, $rt_episode, $rt_year, $rt_director, $rt_cast, $rt_premiere, $rt_film, $rt_repeat, $rt_subtitles, $rt_widescreen, $rt_new_series, $rt_deaf_signed, $rt_black_and_white, $rt_star_rating, $rt_certificate, $rt_genre, $rt_desc, $rt_choice, $rt_date, $rt_start, $rt_stop, $rt_duration_mins) = $fields;
        unset($fields);


        // Include user's code, if any
        if ($this->user_epg_line_fields_split_exists)
            include $this->user_epg_line_fields_split_path;


    	// Roundabout the summer time changeover they include timezone in the title.
        $pattern = '/\((GMT|UTC\+1|BST|UTC)\)\s*/';
        if (preg_match ($pattern, $rt_title , $matches) == 1)
            $rt_title = str_replace('(' . $matches[1] . ')', '', $rt_title);

        if ($this->replacebadchars)
            $rt_title = str_replace(':', '_', $rt_title);       // replace any ':' with '_'

        list($day, $month, $year) = split("\/", $rt_date);
        list($start_hour, $start_minute) = split(":", $rt_start);
        list($stop_hour, $stop_minute) = split(":", $rt_stop);

        $start_time = (int) mktime($start_hour, $start_minute, 0, $month, $day, $year, -1);
        $stop_time = (int) mktime($stop_hour, $stop_minute, 0, $month, $day, $year, -1);
        if ($start_time > $stop_time)
            $stop_time = (int) strtotime("+1 day", $stop_time);

        if (($newepg) && ($stop_time < $this->tstamp))
        {
            if (!$quiet)
                echo "EPG data is old\n";
            return(true);                       // This is old EPG data
        }

        $clipped = "";
        if ($this->svc_limited)
        {
            // This channel is not broadcast 24 hours per day
            $skip = false;
            $svc_start_time = (int) mktime($this->svc_start_hrs, $this->svc_start_mins, 0, $month, $day, $year, -1);
            $svc_stop_time = (int) mktime($this->svc_stop_hrs, $this->svc_stop_mins, 0, $month, $day, $year, -1);

            if ($svc_start_time > $svc_stop_time)
            {
                // If stop time < start time then there are effectively 2 periods to check against
                if ($start_time < $svc_stop_time)
                {
                    if ($stop_time > $svc_stop_time)
                    {
                        $clipped = "(Programme end at " . date("H:i", $stop_time) . " will be missed) ";
                        $stop_time = $svc_stop_time;
                    }
                }
                else if ($start_time >= $svc_start_time)
                {
                    $svc_stop_time = (int) strtotime("+1 day", $svc_stop_time);
                    if ($stop_time > $svc_stop_time)
                    {
                        $clipped = "(Programme end at " . date("H:i", $stop_time) . " will be missed) ";
                        $stop_time = $svc_stop_time;
                    }
                }
                else if ($stop_time > $svc_start_time)
                {
                    $clipped = "(Programme start at " . date("H:i", $start_time) . " will be missed) ";
                    $start_time = $svc_start_time;
                }
                else
                    $skip = true;                           // Skip this program
            }
            else
            {
                if (($start_time < $svc_start_time) && ($stop_time <= $svc_start_time))
                    $skip = true;                           // Skip this program
                else if (($start_time >= $svc_stop_time) && ($stop_time > $svc_stop_time))
                    $skip = true;                           // Skip this program
                else if ($start_time < $svc_start_time)
                {
                    $clipped = "(Programme start at " . date("H:i", $start_time) . " will be missed) ";
                    $start_time = $svc_start_time;
                }
                else if ($stop_time > $svc_stop_time)
                {
                    $clipped = "(Programme end at " . date("H:i", $stop_time) . " will be missed) ";
                    $stop_time = $svc_stop_time;
                }
            }

            if ($skip)
                return(false);
        }
		
		// If no duration, then ignore
		if ($rt_duration_mins == "0")
			return(false);

        // Handle episode and substitle
        if (strlen($rt_episode) > 0)
        {
            $rt_sub_title = $rt_episode;        // ignore subtitle because episode is given

            $pattern = '/^(\d+\/\d+)\s*/';
            $i = preg_match ($pattern, $rt_episode , $matches);
            if ($i > 0)
            {
                $rt_episode = $matches[1];
                $rt_sub_title = substr($rt_sub_title, strlen($matches[1]));
                $rt_sub_title = preg_replace('/^\s*-\s*/', '', $rt_sub_title);
            }
            else
            {
                $rt_sub_title = $rt_episode;      // ignore subtitle because episode is given
                $rt_episode = '';
            }
        }


        // Include user's code, if any
        if ($this->user_pre_mei_line_exists)
            include $this->user_pre_mei_line_path;

		if ($rt_film === 'true')
			$rt_genre = "Film";

        // Construct the string for the line in the mei file
        $mei_str = "";
        $mei_str .= date("YmdHi", $start_time);                                                 // program start time
        $mei_str .= "|" . date("YmdHi", $stop_time);                                            // program stop time
        $mei_str .= "|$this->lcn";                                                              // LCN of channel
        $mei_str .= "|$rt_title";                                                               // program title
        $mei_str .= "|$rt_sub_title";                                                           // secondary title, e.g. episode title
        $mei_str .= "|" . ($rt_repeat === 'true' ? 'TRUE' : '');                                // if this is a repeat
        $mei_str .= "|";                                                                        // if this is first screening of a new show - not in RT data
        $mei_str .= "|$rt_year";                                                                // production date
        $mei_str .= "|" . ($rt_premiere === 'true' ? $rt_genre . ' Premiere' : '');             // if it is a premiere
        $mei_str .= "|";                                                                        // last chance
        $mei_str .= "|$clipped$rt_desc";                                                        // program description
        $mei_str .= "|$rt_genre";                                                               // program genre
        $mei_str .= "|$rt_duration_mins";                                                       // program length
        $mei_str .= "|" . ($rt_widescreen === 'true' ? '16:9' : '');                            // widescreen
        $mei_str .= "|" . ($rt_black_and_white === 'true' ? 'no' : '');                         // whether it is in colour
        $mei_str .= "|" . ($rt_subtitles === 'true' ? 'teletext' : '');                         // whether there are subtitles
        $mei_str .= "|$rt_episode";                                                             // episode number
        $mei_str .= "|" . (($rt_episode != '') ? 'xmltv_ns' : '');                              // episodes start at 1
        $mei_str .= "|" . (($rt_certificate != '') ? 'BBFC' : '');                              // rating system
        $mei_str .= "|$rt_certificate";                                                         // rating
        $mei_str .= "|" . (($rt_star_rating != '') ? $rt_star_rating . '/5' : '');              // star rating
        $mei_str = substr("000000" . strlen($mei_str), -6) . "|$mei_str\r\n";

        // Write the string to the file
        if (fwrite($this->handle, $mei_str) === FALSE)
        {
            if (!$quiet)
                echo "Error writing to $this->filename\n";
            exit(1);
        }

        return(false);                          // Success
    }

}   /* End of MEI class */


function check_time($time_str)
{
    if (strlen($time_str) != 4)
        return(false);

    if (!is_numeric($time_str))
        return(false);

    $hr = substr($time_str, 0, 2);
    if ($hr >= 24)
        return(false);

    $min = substr($time_str, -2);
    if ($min >= 60)
        return(false);

    return(true);
}




/* Start of main code */

// Set user agent to comply with Radio Times requirements
$rt2mei_version = "1.1a";
$ua = "rt2mei $rt2mei_version (http://www.wooders.co.uk/rt2mei), rt2mei@wooders.co.uk";
ini_set("user_agent", $ua);

// Check for version of PHP
$minver = "4.3.5";
if (version_compare(phpversion(), $minver, "<"))
{
    echo "ERROR: You are using PHP version " . phpversion() . "\nPHP version $minver or later is required\n";
    exit(1);
}

$tstamp = (int) mktime(0, 0, 0);                            // 0000 hrs this morning

$dirsep = (preg_match('/Windows/', php_uname('s')) > 0) ? "\\" : "/";

$path_parts = pathinfo(__FILE__);
$script_path = $path_parts['dirname'];
$config_path = $script_path;
$output_path = $script_path;
$temp_path = $script_path;
$cachedata_path = '';
$quiet = false;                                             // default to outputting error messages
$fav_reqd = true;                                           // Assume that we will write a Favourites file
$replacebadchars = false;                                   // Don't replace : in title with _
$wget = false;                                              // Assume wget not used
$newepg = false;                                            // Accept an old EPG
$accempty = false;                                          // Do not accept empty EPG data files
$rt_url = 'http://xmltv.radiotimes.com/xmltv/';				// All data files can be downloaded from here - just append the filename

/* Debug code for checking args passed in
if (!$quiet)
{
    $args = "";
    for ($i = 1; $i < $argc; $i++)
        $args .= "$argv[$i] ";
    echo "Args: $args\n";
}
exit(0);
// */

// Handle parameters
for ($i = 1; $i < $argc; $i++)
{
    switch($argv[$i])
    {
    case "-h":
        echo <<<USAGE
Filename:   rt2mei.php
Version:    $rt2mei_version
Date:       4 March 2009
Author:     Wooders

Usage:
Assuming that PHP is in the PATH then

php <path>rt2mei.php [-- options]

Where options are:
    -h      prints this information

    -q      quiet mode, no error messages output.

    -f      suppresses the writing of a Favourites file

    -r      replace ':' in title with '_'

    -i      followed by a path to the directory where the rt2mei.cfg
            configuration file is located. If omitted then the directory path
            for this script is used.

    -o      followed by a path to the directory where MyStuff.mei and
            rt2mei.mcl will be written. If omitted then the directory
            path for this script is used.

    -w      followed by the full path to the wget executable forces wget to be
            used to grab the EPG data. This is only normally necessary if you
            are behind a corporate firewall or similar. wget is not included in
            rt2mei.

    -t      followed by the full path to the directory where temporary files
            should be written. This is only needed when the -w switch is used.

    -new    rejects EPG data if it contains a program that ends before 0000 hrs
            today

    -mt     does not reject empty EPG data files. I recommend that you DO NOT
            use this option because you may lose EPG data. If you insist on
            using it then be it on your own head, do not come whining to me!


All downloaded data is the copyright of the Radio Times website
(http://www.radiotimes.com).

In accessing this XML feed, you agree that you will only access its contents
for your own personal and non-commercial use and not for any commercial or
other purposes, including advertising or selling any goods or services,
including any third-party software applications available to the general
public.

USAGE;
        exit(1);
        break;

    case "--":
        // Do nothing. Someone has included this on the command line and is using PHP 4
        break;

    case "-q":
        $quiet = true;
        break;

    case "-f":
        $fav_reqd = false;
        break;

    case "-r":
        $replacebadchars = true;
        break;

    case "-i":
        if (++$i < $argc)
            $config_path = $argv[$i];
        break;

    case "-o":
        if (++$i < $argc)
            $output_path = $argv[$i];
        break;

    case "-w":
        $wget = true;
        if (++$i < $argc)
            $wget_exe = $argv[$i];
        break;

    case "-t":
        if (++$i < $argc)
            $temp_path = $argv[$i];
        break;

    case "-new":
        $newepg = true;
        break;

    case "-mt":
        $accempty = true;
        break;

    case "-d":
        if (++$i < $argc)
            $cachedata_path= $argv[$i];
        break;

    case "-s":
        if (++$i < $argc)
            $rt_url = $argv[$i];
        break;

    default:
        echo "ERROR: Invalid argument";
        exit(1);
    }
}

// Strip outer double quotes, replace '\' with '/', then ensure paths ends with '/'
$patterns = array('/^\"/', '/\"$/', '/\\\\/');
$replacements = array('', '', $dirsep);

$script_path = preg_replace($patterns, $replacements, $script_path);
if (substr($script_path, -1) != $dirsep)
    $script_path .= $dirsep;

$config_path = preg_replace($patterns, $replacements, $config_path);
if (substr($config_path, -1) != $dirsep)
    $config_path .= $dirsep;

$temp_path = preg_replace($patterns, $replacements, $temp_path);
if (substr($temp_path, -1) != $dirsep)
    $temp_path .= $dirsep;

$output_path = preg_replace($patterns, $replacements, $output_path);
if (substr($output_path, -1) != $dirsep)
    $output_path .= $dirsep;

// Now we can make some filenames
$config_filename = $config_path . "rt2mei.cfg";             // configuration file
$temp_filename = $temp_path . "epg.tmp";                    // temporary epg data file
$mei_filename = $output_path . "MyStuff.mei";               // output file
$fav_filename = $output_path . "rt2mei.mcl";                // Favourites file

if (!$quiet)
{
    echo "rt2mei.php $rt2mei_version started: " . date("d/m/Y H:i:s", (int) mktime()) . "\n\n";
    echo "Config data will be read from file \"$config_filename\"\n";

    if ($fav_reqd)
        echo "Favourites will be written to file \"$fav_filename\"\n";

    echo "MEI data will be written to file \"$mei_filename\"\n";
    echo "\n";
}

// Filename constants
define("RT2MEI_CHARACTER_REPLACEMENTS",  "rt2mei_character_replacements.php");
define("USER_EPG_LINE_FIELDS_SPLIT", "user_epg_line_fields_split.php");
define("USER_PRE_MEI_LINE", "user_pre_mei_line.php");
define("USER_PRE_CFG_READ", "user_pre_cfg_read.php");
define("USER_PRE_MEI_CREATE", "user_pre_mei_create.php");
define("USER_PRE_CHANNEL_EPG_READ", "user_pre_channel_epg_read.php");
define("USER_POST_CHANNEL_EPG_READ", "user_post_channel_epg_read.php");
define("USER_SUCCESS", "user_success.php");

// Determine if include files exist
$rt2mei_character_replacements_path = $script_path . RT2MEI_CHARACTER_REPLACEMENTS;
$rt2mei_character_replacements_exists = file_exists($rt2mei_character_replacements_path);
$user_epg_line_fields_split_path = $script_path . USER_EPG_LINE_FIELDS_SPLIT;
$user_epg_line_fields_split_exists = file_exists($user_epg_line_fields_split_path);
$user_pre_mei_line_path = $script_path . USER_PRE_MEI_LINE;
$user_pre_mei_line_exists = file_exists($user_pre_mei_line_path);
$user_pre_cfg_read_path = $script_path . USER_PRE_CFG_READ;
$user_pre_cfg_read_exists = file_exists($user_pre_cfg_read_path);
$user_pre_mei_create_path = $script_path . USER_PRE_MEI_CREATE;
$user_pre_mei_create_exists = file_exists($user_pre_mei_create_path);
$user_pre_channel_epg_read_path = $script_path . USER_PRE_CHANNEL_EPG_READ;
$user_pre_channel_epg_read_exists = file_exists($user_pre_channel_epg_read_path);
$user_post_channel_epg_read_path = $script_path . USER_POST_CHANNEL_EPG_READ;
$user_post_channel_epg_read_exists = file_exists($user_post_channel_epg_read_path);
$user_success_path = $script_path . USER_SUCCESS;
$user_success_exists = file_exists($user_success_path);

/*  The format of the config file is:

    <Selected>|<LCN>|<RT id>|<RT Channel Name>|<Service Start Time>|<Service Stop Time>

    where
    <Selected>              is the field used to select which EPG data to
                            download from the Radio Times website.
                            leave blank to not download this data (spaces count
                            as blank).
                            if the -f option is used then any non-blank
                            character(s) will do.
                            if the -f option is not used then this field must
                            be numeric.
                            Channels appear in the Favourites in ascending
                            order of number in this field.
                            Duplicate numbers are not permitted.
    <LCN>                   is an integer and is the Freeview channel number
                            N.B. THIS MAY VARY SLIGHTLY BY REGION.
    <RT id>                 is an integer and is used to derive the file to
                            read from the RT website - DO NOT CHANGE!
    <RT Channel Name>       is the RT's name for the channel - DO NOT CHANGE!
    <Service Start Time>    is the time that broadcasts on the channel start,
                            e.g. 0600
    <Service Stop Time>     is the time that broadcasts on the channel end,
                            e.g. 1800
                            leave both of the above blank for a continuous
                            service.

    N.B. Comments start with '//'
*/


// Include user's code, if any
if ($user_pre_cfg_read_exists)
    include $user_pre_cfg_read_path;


// Read in all the lines in the config file
$config = @file($config_filename);
if ($config === false)
{
    if (!$quiet)
        echo "ERROR: failed to read Configuration file $config_filename\n";
    exit(1);
}


// Validate the config file to ensure that only one RT channel is selected for an LCN
foreach ($config as $line)
{
    $line = rtrim($line);
    $line = preg_replace("@\s*//.*@", "", $line);

    if (strlen($line) > 0)
    {
        if (!$quiet)
            echo "$line\n";

        $fldcnt = 0;
        $fields = explode("|", $line);          // split the line into its various fields
        foreach ($fields as $key => $value)
        {
            if (is_null($value))
                $value = "";
            $fields[$key] = trim($value);       // trim leading and trailing whitespace
            $fldcnt++;
        }

        if ($fldcnt != 6)
        {
            if (!$quiet)
                echo "ERROR: You have $fldcnt fields on the following line in the configuration file. There MUST be 6 fields.\n    '$line'\n";
            exit(1);
        }

        // Write each field into its own variable
        list($selected, $lcn, $rt_id, $channel_name, $svc_start, $svc_stop) = $fields;
        unset($fields);

        // Check the obligatory fields
        if (empty($lcn))
        {
            if (!$quiet)
                echo "ERROR: The LCN field is empty; it must contain a number.\n";
            exit(1);
        }
        else if (!is_numeric($lcn))
        {
            if (!$quiet)
                echo "ERROR: The LCN field does not contain a number.\n";
            exit(1);
        }
        else if (empty($rt_id))
        {
            if (!$quiet)
                echo "ERROR: The RT id field is empty; it must contain a number.\n";
            exit(1);
        }
        else if (!is_numeric($rt_id))
        {
            if (!$quiet)
                echo "ERROR: The RT id field does not contain a number.\n";
            exit(1);
        }
        else if (empty($channel_name))
        {
            if (!$quiet)
                echo "ERROR: The RT Channel Name field is empty; it must contain the name of the channel.\n";
            exit(1);
        }

        $svc_limited = false;
        if (!empty($svc_start) || !empty($svc_stop))
        {
            $svc_limited = true;
            if (!check_time($svc_start))
                $svc_start = "0000";

            if (!check_time($svc_stop))
                $svc_stop = "0000";
        }
        else
            $svc_limited = false;

        if ($selected != '')
        {
            if (isset($worklist[$lcn]))
            {
                if (!$quiet)
                    echo "ERROR: You have selected more than one channel for the same LCN, which are:\n" .
                        $worklist[$lcn]["channel_name"] . " and $channel_name on LCN $lcn\n";
                exit(1);
            }

            $worklist[$lcn] = array("lcn" => $lcn, "rt_id" => $rt_id, "channel_name" => $channel_name,
                                    "svc_limited" => $svc_limited, "svc_start" => $svc_start, "svc_stop" => $svc_stop);

            if ($fav_reqd)
            {
                if (is_numeric($selected))
                {
                    if (isset($favourites[$selected]))
                    {
                        if (!$quiet)
                            echo "ERROR: You have used the same number more than once in the\n'Selected' column of the Configuration file.\n";
                        exit(1);
                    }

                    $favourites[$selected] = array("lcn" => $lcn, "channel_name" => $channel_name);
                }
                else
                {
                    if (!$quiet)
                        echo "ERROR: You have not suppressed the creation of a Favourites file therefore you\nmust only use numbers in the 'Selected' column of the Configuration file.\n";
                    exit(1);
                }
            }
        }
    }
}
unset($config);


if (count($worklist) > 0)
{
    /* The user has selected some channels to be read from the RT website */

    if (!$quiet)
        echo "\n";

    // Write Favourites file before grabbing EPG data
    if ($fav_reqd)
    {
        $handle = @fopen($fav_filename, "w");
        if ($handle === false)
        {
            if (!$quiet)
            {
                echo "ERROR: Failed to open file \"$fav_filename\"\n";
                exit(1);
            }
        }
        else
        {
            if (!$quiet)
                echo "Favourites are:\n";

            ksort($favourites);

            foreach ($favourites as $fav)
            {
                if (@fwrite($handle, $fav["lcn"] . "\r\n") == 0)
                {
                    echo "ERROR: Failed to write to file \"$fav_filename\"\n";
                    exit(1);
                }

                if (!$quiet)
                    echo $fav["lcn"] . "\t" . $fav["channel_name"] . "\n";
            }
            fclose($handle);
        }

        if (!$quiet)
            echo "\n";
    }

    ksort($worklist);           // MEI file will be written in ascending LCN order


    // Include user's code, if any
    if ($user_pre_mei_create_exists)
        include $user_pre_mei_create_path;


    // Grab the EPG data
    $mei = new MEI($mei_filename, $tstamp, $quiet, $replacebadchars, $script_path);   // open the output file
    foreach ($worklist as $workitem)                                    // for each channel selected in the config file
    {
        // Read in the EPG data for this channel
        $mei->set_rt_id($workitem["rt_id"]);
        $mei->set_lcn($workitem["lcn"]);
        $mei->set_rt_channel_name($workitem["channel_name"]);
        $mei->set_svc_hours($workitem["svc_limited"], $workitem["svc_start"], $workitem["svc_stop"]);
        $mei->set_script_files($rt2mei_character_replacements_path, $rt2mei_character_replacements_exists,
                               $user_epg_line_fields_split_path, $user_epg_line_fields_split_exists,
                               $user_pre_mei_line_path, $user_pre_mei_line_exists);

        if (!$quiet)
            echo "... Retrieving data for " . $workitem["channel_name"] . "\n";

        // Try 5 times to read the file
        $epg_read_start_timestamp = (int) mktime();
        for ($file_reads = 5; $file_reads > 0; $file_reads--)
        {
            // Append a random query string to force a bypass of the ISP proxy server
            $rnum = mt_rand();
            $now = (int) mktime();
            $qstr = "$rnum$now";
            $url = $rt_url . $workitem["rt_id"] . ".dat";
			if (strpos($rt_url, "http") === 0)
				$url = $url . "?$qstr";
				
            if ($wget == false)
                $epg_data = @file($url);
            else
            {
                system("$wget_exe -q -O $temp_filename $url", $retval);
                if ($retval == 0)
                {
                    $epg_data = @file($temp_filename);
                    unlink($temp_filename);
                }
                else
                    $epg_data = false;
            }

            if ($epg_data === false)
            {
                // If this is not the last time we have failed then delay 10 secs before trying again
                if ($file_reads > 1)
                {
                    echo "        Failed to read file " . $url . " , trying again in 10 seconds\n";
                    sleep(10);
                }
            }
            else
                break;                                                  // we have data
        }

        if ($epg_data === false)
        {
            // Cannot read the EPG data file so try again later

            if (!$quiet)
                echo "Error whilst reading epg data file $url.\n";
            $mei->destructor();
            exit(2);
        }
		
		if (strlen($cachedata_path) > 0)
		{
            // Write radio times data to cache
			$handle = @fopen($cachedata_path . $workitem["rt_id"] . ".dat", "w");
			if ($handle === false)
			{
				if (!$quiet)
				{
					echo "ERROR: Failed to open file\n";
					exit(1);
				}
			}
			else
			{
				foreach($epg_data as $key => $value)
				{
					if (@fwrite($handle, $value) == 0)
					{
						echo "ERROR: Failed to write to file";
						exit(1);
					}
				}
				fclose($handle);
        	}
		}

        if (count($epg_data) < 3)
        {
            // The EPG data file is empty

            if (!$quiet)
                echo "EPG data for " . $workitem["channel_name"] . " is empty!\n\n";

            if ($accempty == false)
            {
                // Try again later

                $mei->destructor();
                exit(2);
            }
        }

        if (!$quiet)
            echo "... Processing " . count($epg_data) . " lines of data for " . $workitem["channel_name"] . "\n";

        $epg_process_start_timestamp = (int) mktime();

        // Include user's code, if any
        if ($user_pre_channel_epg_read_exists)
            include $user_pre_channel_epg_read_path;


        // Process each line of EPG data
        foreach ($epg_data as $program_data)
        {
            $oldepg = $mei->process_line($program_data, $newepg);
            if ($oldepg)
            {
                // The EPG data is old so try again later

                $mei->destructor();
                exit(2);
            }
        }


        // Include user's code, if any
        if ($user_post_channel_epg_read_exists)
            include $user_post_channel_epg_read_path;

        $epg_process_time = (int) mktime() - $epg_process_start_timestamp;
        echo "... " . $workitem["channel_name"] . " downloaded in " . ($epg_process_start_timestamp - $epg_read_start_timestamp) . " secs and processed in " . $epg_process_time . " secs\n\n";

        if ($epg_process_time < 3)
            sleep(3 - $epg_process_time);   // A little kindness to the RT webserver
    }
    $mei->destructor();

    if (!$quiet)
        echo "\nRetrieved data for " . count($worklist) . " channels\n\n";


    // Include user's code, if any
    if ($user_success_exists)
        include $user_success_path;


    // All done
    exit(0);
}
else
{
    if (!$quiet)
        echo "ERROR: No channels selected.\n";
    exit(1);
}

/* End of file*/
?>
